/*
 * crypt.c - password encryption
 */

static char rcsid[] = "$Header: RCS/crypt.c 1.5 91/01/18 20:45:06 kmori Exp $";

#include <string.h>


typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned long ulong;




static char Keys[16][8];

void qsetkey(char *key)
{
	static char nshifts[]
		= { 1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1 };
#define	C	28
#define	D	0
	static char pc1[] = {
		C + 20, C + 12, C + 4, D + 0, D + 4, D + 12, D + 20,
		C + 21, C + 13, C + 5, D + 1, D + 5, D + 13, D + 21,
		C + 22, C + 14, C + 6, D + 2, D + 6, D + 14, D + 22,
		C + 23, C + 15, C + 7, D + 3, D + 7, D + 15, D + 23,
		C + 24, C + 16, C + 8, C + 0, D + 8, D + 16, D + 24,
		C + 25, C + 17, C + 9, C + 1, D + 9, D + 17, D + 25,
		C + 26, C + 18, C +10, C + 2, D +10, D + 18, D + 26,
		C + 27, C + 19, C +11, C + 3, D +11, D + 19, D + 27
	};
	static char pc2_c[] = {
		14, 11, 17,  4, 27, 23,
		25,  0, 13, 22,  7, 18,
		 5,  9, 16, 24,  2, 20,
		12, 21,  1,  8, 15, 26
	};
	static char pc2_d[] = {
		15,  4, 25, 19,  9,  1,
		26, 16,  5, 11, 23,  8,
		12,  7, 17,  0, 22,  3,
		10, 14,  6, 20, 27, 24,
	};
	ulong c, d;
	uchar b, *p;
	int i, j;

	/* PC-1 (Permuted Choice 1) */
	c = d = 0;
	p = pc1;
	for (i = 0; i < 8; i++) {
		b = *key++ << 1;
		for (j = 0; j < 7; j++) {
			if (b & 0x80) {
				if (*p >= C)
					c |= 1L << (*p - C);
				else
					d |= 1L << *p;
			}
			p++;
			b <<= 1;
		}
	}

	for (i = 0; i < 16; i++) {
		uchar *k, mask;
		uint n = nshifts[i];

		c = (c << n | c >> (28 - n)) & 0x0fffffff;
		d = (d << n | d >> (28 - n)) & 0x0fffffff;
		k = Keys[i];
		/* PC-2 (Permuted Choice 2) */
		p = pc2_c;
		for (j = 0; j < 4; j++) {
			*k = 0;
			for (mask = 0x20; mask > 0; mask >>= 1) {
				if ((c >> *p++) & 1)
					*k |= mask;
			}
			k++;
		}
		p = pc2_d;
		for (j = 0; j < 4; j++) {
			*k = 0;
			for (mask = 0x20; mask > 0; mask >>= 1) {
				if ((d >> *p++) & 1)
					*k |= mask;
			}
			k++;
		}
	}
}



void initial_transpose(char *d, char *s)
{
	char c, b;

	memset(d, 0, 8);
	c = *s++;
	b = 1;
	if (c & 128)	d[4] = b;
	if (c & 64)	d[0] = b;
	if (c & 32)	d[5] = b;
	if (c & 16)	d[1] = b;
	if (c & 8)	d[6] = b;
	if (c & 4)	d[2] = b;
	if (c & 2)	d[7] = b;
	if (c & 1)	d[3] = b;
	c = *s++;
	b <<= 1;
	if (c & 128)	d[4] |= b;
	if (c & 64)	d[0] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[1] |= b;
	if (c & 8)	d[6] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[7] |= b;
	if (c & 1)	d[3] |= b;
	c = *s++;
	b <<= 1;
	if (c & 128)	d[4] |= b;
	if (c & 64)	d[0] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[1] |= b;
	if (c & 8)	d[6] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[7] |= b;
	if (c & 1)	d[3] |= b;
	c = *s++;
	b <<= 1;
	if (c & 128)	d[4] |= b;
	if (c & 64)	d[0] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[1] |= b;
	if (c & 8)	d[6] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[7] |= b;
	if (c & 1)	d[3] |= b;
	c = *s++;
	b <<= 1;
	if (c & 128)	d[4] |= b;
	if (c & 64)	d[0] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[1] |= b;
	if (c & 8)	d[6] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[7] |= b;
	if (c & 1)	d[3] |= b;
	c = *s++;
	b <<= 1;
	if (c & 128)	d[4] |= b;
	if (c & 64)	d[0] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[1] |= b;
	if (c & 8)	d[6] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[7] |= b;
	if (c & 1)	d[3] |= b;
	c = *s++;
	b <<= 1;
	if (c & 128)	d[4] |= b;
	if (c & 64)	d[0] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[1] |= b;
	if (c & 8)	d[6] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[7] |= b;
	if (c & 1)	d[3] |= b;
	c = *s++;
	b <<= 1;
	if (c & 128)	d[4] |= b;
	if (c & 64)	d[0] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[1] |= b;
	if (c & 8)	d[6] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[7] |= b;
	if (c & 1)	d[3] |= b;
}


void final_transpose(char *d, char *s)
{
	char c, b;

	memset(d, 0, 8);
	c = *s++;
	b = 64;
	if (c & 128)	d[7] = b;
	if (c & 64)	d[6] = b;
	if (c & 32)	d[5] = b;
	if (c & 16)	d[4] = b;
	if (c & 8)	d[3] = b;
	if (c & 4)	d[2] = b;
	if (c & 2)	d[1] = b;
	if (c & 1)	d[0] = b;
	c = *s++;
	b = 16;
	if (c & 128)	d[7] |= b;
	if (c & 64)	d[6] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[4] |= b;
	if (c & 8)	d[3] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[1] |= b;
	if (c & 1)	d[0] |= b;
	c = *s++;
	b = 4;
	if (c & 128)	d[7] |= b;
	if (c & 64)	d[6] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[4] |= b;
	if (c & 8)	d[3] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[1] |= b;
	if (c & 1)	d[0] |= b;
	c = *s++;
	b = 1;
	if (c & 128)	d[7] |= b;
	if (c & 64)	d[6] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[4] |= b;
	if (c & 8)	d[3] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[1] |= b;
	if (c & 1)	d[0] |= b;
	c = *s++;
	b = 128;
	if (c & 128)	d[7] |= b;
	if (c & 64)	d[6] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[4] |= b;
	if (c & 8)	d[3] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[1] |= b;
	if (c & 1)	d[0] |= b;
	c = *s++;
	b = 32;
	if (c & 128)	d[7] |= b;
	if (c & 64)	d[6] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[4] |= b;
	if (c & 8)	d[3] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[1] |= b;
	if (c & 1)	d[0] |= b;
	c = *s++;
	b = 8;
	if (c & 128)	d[7] |= b;
	if (c & 64)	d[6] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[4] |= b;
	if (c & 8)	d[3] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[1] |= b;
	if (c & 1)	d[0] |= b;
	c = *s++;
	b = 2;
	if (c & 128)	d[7] |= b;
	if (c & 64)	d[6] |= b;
	if (c & 32)	d[5] |= b;
	if (c & 16)	d[4] |= b;
	if (c & 8)	d[3] |= b;
	if (c & 4)	d[2] |= b;
	if (c & 2)	d[1] |= b;
	if (c & 1)	d[0] |= b;
}


static char Esalt[2];


static ulong f(int i, ulong r)
{
	static uchar sboxes[8][64] = {
	{
	14,  0,  4, 15, 13,  7,  1,  4,  2, 14, 15,  2, 11, 13,  8,  1,
	 3, 10, 10,  6,  6, 12, 12, 11,  5,  9,  9,  5,  0,  3,  7,  8,
	 4, 15,  1, 12, 14,  8,  8,  2, 13,  4,  6,  9,  2,  1, 11,  7,
	15,  5, 12, 11,  9,  3,  7, 14,  3, 10, 10,  0,  5,  6,  0, 13,
	},
	{
	15,  3,  1, 13,  8,  4, 14,  7,  6, 15, 11,  2,  3,  8,  4, 14,
	 9, 12,  7,  0,  2,  1, 13, 10, 12,  6,  0,  9,  5, 11, 10,  5,
	 0, 13, 14,  8,  7, 10, 11,  1, 10,  3,  4, 15, 13,  4,  1,  2,
	 5, 11,  8,  6, 12,  7,  6, 12,  9,  0,  3,  5,  2, 14, 15,  9,
	},
	{
	10, 13,  0,  7,  9,  0, 14,  9,  6,  3,  3,  4, 15,  6,  5, 10,
	 1,  2, 13,  8, 12,  5,  7, 14, 11, 12,  4, 11,  2, 15,  8,  1,
	13,  1,  6, 10,  4, 13,  9,  0,  8,  6, 15,  9,  3,  8,  0,  7,
	11,  4,  1, 15,  2, 14, 12,  3,  5, 11, 10,  5, 14,  2,  7, 12,
	},
	{
	 7, 13, 13,  8, 14, 11,  3,  5,  0,  6,  6, 15,  9,  0, 10,  3,
	 1,  4,  2,  7,  8,  2,  5, 12, 11,  1, 12, 10,  4, 14, 15,  9,
	10,  3,  6, 15,  9,  0,  0,  6, 12, 10, 11,  1,  7, 13, 13,  8,
	15,  9,  1,  4,  3,  5, 14, 11,  5, 12,  2,  7,  8,  2,  4, 14,
	},
	{
	 2, 14, 12, 11,  4,  2,  1, 12,  7,  4, 10,  7, 11, 13,  6,  1,
	 8,  5,  5,  0,  3, 15, 15, 10, 13,  3,  0,  9, 14,  8,  9,  6,
	 4, 11,  2,  8,  1, 12, 11,  7, 10,  1, 13, 14,  7,  2,  8, 13,
	15,  6,  9, 15, 12,  0,  5,  9,  6, 10,  3,  4,  0,  5, 14,  3,
	},
	{
	12, 10,  1, 15, 10,  4, 15,  2,  9,  7,  2, 12,  6,  9,  8,  5,
	 0,  6, 13,  1,  3, 13,  4, 14, 14,  0,  7, 11,  5,  3, 11,  8,
	 9,  4, 14,  3, 15,  2,  5, 12,  2,  9,  8,  5, 12, 15,  3, 10,
	 7, 11,  0, 14,  4,  1, 10,  7,  1,  6, 13,  0, 11,  8,  6, 13,
	},
	{
	 4, 13, 11,  0,  2, 11, 14,  7, 15,  4,  0,  9,  8,  1, 13, 10,
	 3, 14, 12,  3,  9,  5,  7, 12,  5,  2, 10, 15,  6,  8,  1,  6,
	 1,  6,  4, 11, 11, 13, 13,  8, 12,  1,  3,  4,  7, 10, 14,  7,
	10,  9, 15,  5,  6,  0,  8, 15,  0, 14,  5,  2,  9,  3,  2, 12,
	},
	{
	13,  1,  2, 15,  8, 13,  4,  8,  6, 10, 15,  3, 11,  7,  1,  4,
	10, 12,  9,  5,  3,  6, 14, 11,  5,  0,  0, 14, 12,  9,  7,  2,
	 7,  2, 11,  1,  4, 14,  1,  7,  9,  4, 12, 10, 14,  8,  2, 13,
	 0, 15,  6, 12, 10,  9, 13,  0, 15,  3,  3,  5,  5,  6,  8, 11,
	},
	};
	static ulong ptbl[8][16] = {
	{
	0x00000000, 0x00000002, 0x00000200, 0x00000202,
	0x00008000, 0x00008002, 0x00008200, 0x00008202,
	0x00800000, 0x00800002, 0x00800200, 0x00800202,
	0x00808000, 0x00808002, 0x00808200, 0x00808202,
	},
	{
	0x00000000, 0x00004000, 0x40000000, 0x40004000,
	0x00000010, 0x00004010, 0x40000010, 0x40004010,
	0x00080000, 0x00084000, 0x40080000, 0x40084000,
	0x00080010, 0x00084010, 0x40080010, 0x40084010,
	},
	{
	0x00000000, 0x04000000, 0x00000004, 0x04000004,
	0x00010000, 0x04010000, 0x00010004, 0x04010004,
	0x00000100, 0x04000100, 0x00000104, 0x04000104,
	0x00010100, 0x04010100, 0x00010104, 0x04010104,
	},
	{
	0x00000000, 0x80000000, 0x00400000, 0x80400000,
	0x00001000, 0x80001000, 0x00401000, 0x80401000,
	0x00000040, 0x80000040, 0x00400040, 0x80400040,
	0x00001040, 0x80001040, 0x00401040, 0x80401040,
	},
	{
	0x00000000, 0x20000000, 0x00000080, 0x20000080,
	0x00040000, 0x20040000, 0x00040080, 0x20040080,
	0x01000000, 0x21000000, 0x01000080, 0x21000080,
	0x01040000, 0x21040000, 0x01040080, 0x21040080,
	},
	{
	0x00000000, 0x00002000, 0x00200000, 0x00202000,
	0x00000008, 0x00002008, 0x00200008, 0x00202008,
	0x10000000, 0x10002000, 0x10200000, 0x10202000,
	0x10000008, 0x10002008, 0x10200008, 0x10202008,
	},
	{
	0x00000000, 0x02000000, 0x00000400, 0x02000400,
	0x00100000, 0x02100000, 0x00100400, 0x02100400,
	0x00000001, 0x02000001, 0x00000401, 0x02000401,
	0x00100001, 0x02100001, 0x00100401, 0x02100401,
	},
	{
	0x00000000, 0x00000800, 0x00020000, 0x00020800,
	0x00000020, 0x00000820, 0x00020020, 0x00020820,
	0x08000000, 0x08000800, 0x08020000, 0x08020800,
	0x08000020, 0x08000820, 0x08020020, 0x08020820,
	},
	};
	uchar frags[8];
	uchar c;
	uchar *key;
	ulong v;
	int j;

	/* Transpose E */
	r = r << 1 | r >> 31;
	for (j = 8; --j >= 0; ) {
		frags[j] = (uchar)r & 0x3f;
		r = r >> 4 | r << 28;
	}

	/* Salt */
	c = frags[4] & ~Esalt[0] | frags[0] & Esalt[0];
	frags[0] = frags[0] & ~Esalt[0] | frags[4] & Esalt[0];
	frags[4] = c;
	c = frags[5] & ~Esalt[1] | frags[1] & Esalt[1];
	frags[1] = frags[1] & ~Esalt[1] | frags[5] & Esalt[1];
	frags[5] = c;

	/* Xor, S-boxes and P-transpose */
	key = Keys[i];
	v = 0;
	for (j = 0; j < 8; j++)
		v |= ptbl[j][sboxes[j][key[j] ^ frags[j]]];
	return (v);
}



void qencrypt(char *block, int edflag)
{
	uchar w[8];
	ulong l, r;
	int i;

	initial_transpose(w, block);
	l =   (ulong)w[0] << 24
	    | (ulong)w[1] << 16
	    | (ulong)w[2] << 8
	    | (ulong)w[3];
	r =   (ulong)w[4] << 24
	    | (ulong)w[5] << 16
	    | (ulong)w[6] << 8
	    | (ulong)w[7];
	for (i = 0; i < 16; i++) {
		ulong pl;

		pl = l;
		l = r;
		r = pl ^ f(edflag ? 15 - i : i, r);
	}
	w[0] = r >> 24;
	w[1] = r >> 16;
	w[2] = r >> 8;
	w[3] = r;
	w[4] = l >> 24;
	w[5] = l >> 16;
	w[6] = l >> 8;
	w[7] = l;
	final_transpose(block, w);
}



char *crypt(char *pw, char *salt)
{
	static char text[14];
	uchar key[8];
	uchar cipher[9];
	uchar *p;
	int i;

	/* Set salt */
	for (i = 0; i < 2; i++) {
		char c = salt[i];
		char d;
		int j;

		if (c > 'Z')
			c -= 6 + 7 + '.';
		else if (c > '9')
			c -= 7 + '.';
		else
			c -= '.';
		/* d = bit-reverse(c); */
		d = 0;
		for (j = 0; j < 6; j++) {
			d = d << 1 | (c & 1);
			c >>= 1;
		}
		Esalt[i] = d;
	}

	/* Copy password to key with zero-fill */
	memset(key, 0, 8);
	strncpy(key, pw, 8);
	qsetkey(key);

	/* Encrypt eight 0 block 25 times */
	memset(cipher, 0, 9);
	for (i = 0; i < 25; i++)
		qencrypt(cipher, 0);

	Esalt[0] = Esalt[1] = 0;

	/* Convert binary cipher to visible ASCII text */
	text[0] = salt[0];
	text[1] = salt[1];
	p = text + 2;
	for (i = 0; i < 64; i += 6) {
		uint w;

		w = cipher[i / 8] << 8 | cipher[i / 8 + 1];
		w >>= 10 - i % 8;
		w &= 0x3f;
		w += '.';
		if (w > '9')
			w += 7;
		if (w > 'Z')
			w += 6;
		*p++ = w;
	}
	*p = '\0';
	return (text);
}
